'From Squeak4.1 of 20 May 2010 [latest update: #10154] on 25 May 2010 at 6:12:38 am'!
Collection subclass: #WeakFinalizationRegistry
	instanceVariableNames: 'list valueDictionary sema '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Collections-Weak'!
WeakRegistryTest subclass: #WeakFinalizationRegistryTest
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'CollectionsTests-Weak'!

!Object methodsFor: 'finalization' stamp: 'Igor.Stasenko 5/25/2010 04:59'!
hasMultipleExecutors
	"All objects, except ObjectFinalizerCollection instances should answer false to this message"
	^ false! !


!ObjectFinalizerCollection methodsFor: 'testing' stamp: 'Igor.Stasenko 5/25/2010 04:58'!
hasMultipleExecutors
	^ true! !


!WeakFinalizationRegistry methodsFor: 'printing' stamp: 'Igor.Stasenko 5/25/2010 05:13'!
printElementsOn: aStream
	sema ifNil: [^super printElementsOn: aStream].
	aStream nextPutAll: '(<this WeakRegistry is locked>)'! !

!WeakFinalizationRegistry methodsFor: 'printing' stamp: 'Igor.Stasenko 5/25/2010 05:21'!
remove: oldObject ifAbsent: exceptionBlock
	"Remove oldObject as one of the receiver's elements."
	
	oldObject ifNil: [ ^nil ].
	^(self protected: [ valueDictionary removeKey: oldObject ifAbsent: nil ])
		ifNil: [ exceptionBlock value ]! !

!WeakFinalizationRegistry methodsFor: 'printing' stamp: 'Igor.Stasenko 5/25/2010 05:16'!
removeAll
	"See super"
	
	self protected:[
		valueDictionary removeAll.
	].! !

!WeakFinalizationRegistry methodsFor: 'enumerating' stamp: 'Igor.Stasenko 5/25/2010 05:11'!
do: aBlock
	^self protected: [
		valueDictionary keysDo: aBlock.
	].
! !

!WeakFinalizationRegistry methodsFor: 'adding' stamp: 'Igor.Stasenko 5/25/2010 05:30'!
add: anObject executor: anExecutor

	self protected: [ | finItem |
		finItem := valueDictionary at: anObject ifAbsentPut: [
			WeakFinalizerItem new list: list object: anObject ].
		finItem add: anExecutor ].
	^ anObject
! !

!WeakFinalizationRegistry methodsFor: 'finalization' stamp: 'Igor.Stasenko 5/25/2010 05:43'!
finalizeValues
	"Finalize any values, which happen to stocked in our list, due to some weak references become garbage"
	
	| finalizer |

	self protected: [ finalizer := list swapWithNil ].

	"We don't need to proted a following loop from concurrent access,
	because at the moment we're finalizing values, 
	only we can access this list of finalizers, because valueDictionary already see them
	as an unused slots, because they're associated with key == nil"
	
	[ finalizer notNil ] whileTrue: [
		| next |
		next := finalizer next.
		finalizer finalizeValues.
		finalizer := next
	].
! !

!WeakFinalizationRegistry methodsFor: 'initialize-release' stamp: 'Igor.Stasenko 5/25/2010 04:38'!
initialize
	valueDictionary := WeakIdentityKeyDictionary new.
	list := WeakFinalizationList new.
	sema := Semaphore forMutualExclusion.! !

!WeakFinalizationRegistry methodsFor: 'accessing' stamp: 'Igor.Stasenko 5/25/2010 05:48'!
size
	^ self protected: [valueDictionary slowSize]! !

!WeakFinalizationRegistry methodsFor: 'private' stamp: 'Igor.Stasenko 5/25/2010 05:34'!
protected: aBlock
	"Execute aBlock protected by the accessLock"

	^ sema
		critical: aBlock
		ifError: [ :msg :rcvr |
		rcvr error: msg ] ! !

!WeakFinalizationRegistry methodsFor: 'private' stamp: 'Igor.Stasenko 5/25/2010 05:35'!
species
	^Set! !


!WeakFinalizationRegistry class methodsFor: 'instance creation' stamp: 'Igor.Stasenko 5/25/2010 05:31'!
new: n
	^ self new! !


!WeakFinalizerItem methodsFor: 'accessing' stamp: 'Igor.Stasenko 5/25/2010 06:02'!
add: newExecutor

	executor 
		ifNil: [ executor := newExecutor ]
		ifNotNil: [
			executor hasMultipleExecutors
				ifTrue: [ executor add: newExecutor]
				ifFalse: [ executor := ObjectFinalizerCollection with: executor with: newExecutor ]
		]! !

!WeakFinalizerItem methodsFor: 'initialize-release' stamp: 'Igor.Stasenko 5/25/2010 05:29'!
list: weakFinalizationList object: anObject
	self assert: (weakFinalizationList class == WeakFinalizationList).
	list := weakFinalizationList.
	self at: 1 put: anObject.
! !

!WeakFinalizerItem methodsFor: 'copying' stamp: 'Igor.Stasenko 5/25/2010 05:33'!
postCopy
	executor hasMultipleFinalizers ifTrue: [ executor := executor copy ].! !

!WeakFinalizerItem methodsFor: 'finalizing' stamp: 'Igor.Stasenko 5/25/2010 05:39'!
finalizeValues
	" cleanup the receiver, so it could be reused "
	| ex |
	ex := executor.
	executor := nil.
	next := nil.
	ex finalize.! !


!WeakRegistryTest methodsFor: 'tests' stamp: 'Igor.Stasenko 5/25/2010 04:45'!
testFinalization

	| w finalized block object |
	w := self registryClass new: 1.
	finalized := false.
	block := [ :v | finalized := v ].
	object := ObjectFinalizer "an object that responds to #finalize"
		receiver: block
		selector: #value:
		argument: true.
	w add: object.
	object := nil. "let it go"
	Smalltalk garbageCollect. "finalize it"

	"This is an odd issue. It seems that in some situations the finalization
	process doesn't run 'in time' for the isEmpty assertion below to succeed.
	This really *shouldn't* happen since isEmpty is full of real sends and
	there ought to be an interrupt check in there somewhere. However,
	since there are no real-time guarantees about finalization, it's fair to
	just wait a little to ensure that the finalization process has been run."
	(Delay forMilliseconds: 100) wait.

	self assert: w isEmpty.
	self assert: finalized! !

!WeakRegistryTest methodsFor: 'tests' stamp: 'Igor.Stasenko 5/25/2010 04:45'!
testFinalizationWithMultipleFinalizersPerObject

	| object registry counter |
	registry := self registryClass new.
	object := Object new.
	counter := 0.
	5 timesRepeat: [
		registry add: object executor: (ObjectFinalizer
			receiver: [ counter := counter + 1 ]
			selector: #value) ].
	self assert: registry size = 1.
	object := nil. 
	Smalltalk garbageCollect.
	registry finalizeValues.
	self assert: registry isEmpty.
	self assert: counter = 5
! !

!WeakRegistryTest methodsFor: 'tests' stamp: 'Igor.Stasenko 5/25/2010 04:46'!
testGrowingByAHighPriorityProcessDoesntLeak

	| w finalized block object executor semaphore |
	w := self registryClass new: 1.
	finalized := false.
	block := [ :v | finalized := v ].
	object := Object new.
	executor := ObjectFinalizer
		receiver: block
		selector: #value:
		argument: true.
	w add: object executor: executor.
	semaphore := Semaphore new.
	[ 
		object := nil. "let it go"
		w addAll: (1 to: 1000). "force growing"
		semaphore signal ] 
			forkAt: WeakArray runningFinalizationProcess priority + 1.
	semaphore wait.
	Smalltalk garbageCollect. "finalize it"
	self assert: w size = 1000.
	self assert: finalized! !

!WeakRegistryTest methodsFor: 'tests' stamp: 'Igor.Stasenko 5/25/2010 04:46'!
testGrowingDoesntLeak

	| w finalized block object executor |
	w := self registryClass new: 1.
	finalized := false.
	block := [ :v | finalized := v ].
	object := Object new.
	executor := ObjectFinalizer
		receiver: block
		selector: #value:
		argument: true.
	w add: object executor: executor.
	object := nil. "let it go"
	w addAll: (1 to: 1000). "force growing"
	Smalltalk garbageCollect. "finalize it"
	self assert: w size = 1000.
	self assert: finalized! !

!WeakRegistryTest methodsFor: 'tests' stamp: 'Igor.Stasenko 5/25/2010 04:46'!
testRemovingByAHighPriorityProcessDoesntLeak

	| w finalized block hash object executor semaphore |
	w := self registryClass new: 1.
	finalized := false.
	block := [ :v | finalized := v ].
	object := Object new.
	executor := ObjectFinalizer
		receiver: block
		selector: #value:
		argument: true.
	hash := object hash.
	w add: hash.
	w add: object executor: executor.
	semaphore := Semaphore new.
	[ 
		object := nil. "let it go"
		w remove: hash.
		semaphore signal ] 
			forkAt: WeakArray runningFinalizationProcess priority + 1.
	semaphore wait.
	Smalltalk garbageCollect. "finalize it"
	self assert: w isEmpty.
	self assert: finalized! !

!WeakRegistryTest methodsFor: 'tests' stamp: 'Igor.Stasenko 5/25/2010 04:46'!
testRemovingDoesntLeak

	| w finalized block hash object executor |
	w := self registryClass new: 1.
	finalized := false.
	block := [ :v | finalized := v ].
	object := Object new.
	executor := ObjectFinalizer
		receiver: block
		selector: #value:
		argument: true.
	hash := object hash.
	w add: hash.
	w add: object executor: executor.
	object := nil. "let it go"
	w remove: hash.
	Smalltalk garbageCollect. "finalize it"
	self assert: w isEmpty.
	self assert: finalized! !

!WeakRegistryTest methodsFor: 'registry' stamp: 'Igor.Stasenko 5/25/2010 04:44'!
registryClass
	^ WeakRegistry! !


!WeakFinalizationRegistryTest methodsFor: 'as yet unclassified' stamp: 'Igor.Stasenko 5/25/2010 05:45'!
registryClass
	^ WeakFinalizationRegistry! !


!WeakRegistryTest reorganize!
('tests' testFinalization testFinalizationWithMultipleFinalizersPerObject testGrowingByAHighPriorityProcessDoesntLeak testGrowingDoesntLeak testRemovingByAHighPriorityProcessDoesntLeak testRemovingDoesntLeak)
('registry' registryClass)
!


!WeakFinalizerItem reorganize!
('accessing' add: clear executor list list: next object)
('initialize-release' list:object: list:object:executor:)
('copying' copyWithList: postCopy)
('finalizing' finalizeValues)
!


!WeakFinalizationRegistry class reorganize!
('instance creation' new new:)
!

Collection subclass: #WeakFinalizationRegistry
	instanceVariableNames: 'list valueDictionary sema'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Collections-Weak'!

!WeakFinalizationRegistry reorganize!
('printing' printElementsOn: remove:ifAbsent: removeAll)
('enumerating' do:)
('adding' add: add:executor:)
('copying' postCopy)
('finalization' finalizeValues)
('initialize-release' initialize)
('accessing' keys size)
('private' protected: species)
!


!ObjectFinalizerCollection reorganize!
('finalization' finalize)
('testing' hasMultipleExecutors)
!

